#!/usr/bin/env python

"""
A simple (and sane) text indexing library.

Copyright (C) 2009 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
"""

from os import mkdir             # to determine whether to create indexes
from os.path import exists, join
from os.path import commonprefix # to find common string prefixes
from bisect import bisect_right  # to find terms in the dictionary index
import bz2, zlib                 # for field compression

# Constants.

INTERVAL = 100

compressors = [("b", bz2.compress), ("z", zlib.compress)]
decompressors = {"b" : bz2.decompress, "z" : zlib.decompress}

# Foundation classes.

class File:

    "A basic file abstraction."

    def __init__(self, f):
        self.f = f
        self.reset()

    def reset(self):
        pass

    def close(self):
        if self.f is not None:
            self.f.close()
            self.f = None

class FileWriter(File):

    "Writing basic data types to files."

    def write_number(self, number):

        "Write 'number' to the file using a variable length encoding."

        # Negative numbers are not supported.

        if number < 0:
            raise ValueError, "Number %r is negative." % number

        # Special case: one byte containing zero.

        elif number == 0:
            self.f.write(chr(0))
            return

        # Write the number from least to most significant digits.

        bytes = []

        while number != 0:
            lsd = number & 127
            number = number >> 7
            if number != 0:
                lsd |= 128
            bytes.append(chr(lsd))

        record = "".join(bytes)
        self.f.write(record)

    def write_string(self, s, compress=0):

        """
        Write 's' to the file, recording its length and compressing the string
        if 'compress' is set to a true value.
        """

        # Convert Unicode objects to strings.

        if isinstance(s, unicode):
            s = s.encode("utf-8")

        # Compress the string if requested.

        if compress:
            for flag, fn in compressors:
                cs = fn(s)

                # Take the first string shorter than the original.

                if len(cs) < len(s):
                    s = cs
                    break
            else:
                flag = "-"

            # Record whether compression was used.

            self.f.write(flag)

        # Write the length of the data before the data itself.

        length = len(s)
        self.write_number(length)
        self.f.write(s)

class FileReader(File):

    "Reading basic data types from files."

    def read_number(self):

        "Read a number from the file."

        # Read each byte, adding it to the number.

        shift = 0
        number = 0
        more = 1

        while more:
            byte = self.f.read(1)
            if not byte:
                raise EOFError

            csd = ord(byte)
            more = csd & 128 != 0
            if more:
                csd &= 127
            number += (csd << shift)
            shift += 7

        return number

    def read_string(self, decompress=0):

        """
        Read a string from the file, decompressing the stored data if
        'decompress' is set to a true value.
        """

        # Decompress the data if requested.

        if decompress:
            flag = self.f.read(1)
        else:
            flag = "-"

        length = self.read_number()
        s = self.f.read(length)

        # Perform decompression if applicable.

        if flag != "-":
            fn = decompressors[flag]
            s = fn(s)

        # Convert strings to Unicode objects.

        return unicode(s, "utf-8")

# Specific classes for storing term and position information.

class PositionWriter(FileWriter):

    "Writing position information to files."

    def reset(self):
        self.last_docnum = 0

    def write_positions(self, docnum, positions):

        "Write for the document 'docnum' the given 'positions'."

        if docnum < self.last_docnum:
            raise ValueError, "Document number %r is less than previous number %r." % (docnum, self.last_docnum)

        # Write the document number delta.

        self.write_number(docnum - self.last_docnum)

        # Write the number of positions.

        self.write_number(len(positions))

        # Make sure that the positions are sorted.

        positions.sort()

        # Write the position deltas.

        last = 0
        for position in positions:
            pos = position - last
            self.write_number(pos)
            last = position

        self.last_docnum = docnum

    def write_all_positions(self, doc_positions):

        """
        Write all 'doc_positions' - a collection of tuples of the form (document
        number, position list) - to the file, returning a tuple containing the
        offset at which they were stored together with the frequency (number of
        positions) for the term involved.
        """

        # Reset the writer and record the current file offset.

        self.reset()
        offset = self.f.tell()

        # Write the number of documents.

        self.write_number(len(doc_positions))

        # Write the positions.

        frequency = 0

        for docnum, positions in doc_positions:
            self.write_positions(docnum, positions)
            frequency += len(positions)

        return offset, frequency

class PositionReader(FileReader):

    "Reading position information from files."

    def reset(self):
        self.last_docnum = 0

    def read_positions(self):

        "Read positions, returning a document number and a list of positions."

        # Read the document number delta and add it to the last number.

        self.last_docnum += self.read_number()

        # Read the number of positions.

        npositions = self.read_number()

        # Read the position deltas, adding each previous position to get the
        # appropriate collection of absolute positions.

        i = 0
        last = 0
        positions = []

        while i < npositions:
            last += self.read_number()
            positions.append(last)
            i += 1

        return self.last_docnum, positions

    def read_all_positions(self, offset):

        """
        Read all positions from 'offset', seeking to that position in the file
        before reading.
        """

        self.reset()
        self.f.seek(offset)

        # Read the number of documents.

        ndocuments = self.read_number()

        # Read all records.

        i = 0
        doc_positions = []

        while i < ndocuments:
            doc_positions.append(self.read_positions())
            i += 1

        return doc_positions

class TermWriter(FileWriter):

    "Writing term information to files."

    def reset(self):
        self.last_term = ""
        self.last_offset = 0

    def write_term(self, term, offset, frequency):

        """
        Write the given 'term', its position file 'offset', and its 'frequency'
        to the term information file. Return the offset after the term
        information was written to the file.
        """

        # Too long terms are not currently supported.

        if len(term) > 255:
            raise ValueError, "Term %r is too long." % term

        # Write the prefix length and term suffix.

        common = len(commonprefix([self.last_term, term]))
        suffix = term[common:]

        self.write_number(common)
        self.write_string(suffix)

        # Write the offset delta.

        self.write_number(offset - self.last_offset)

        # Write the frequency.

        self.write_number(frequency)

        self.last_term = term
        self.last_offset = offset

        return self.f.tell()

class TermReader(FileReader):

    "Reading term information from files."

    def reset(self):
        self.last_term = ""
        self.last_offset = 0

    def read_term(self):

        """
        Read a term, its position file offset, and its frequency from the term
        information file.
        """

        # Read the prefix length and term suffix.

        common = self.read_number()
        suffix = self.read_string()

        self.last_term = self.last_term[:common] + suffix

        # Read the offset delta.

        self.last_offset += self.read_number()

        # Read the frequency.

        frequency = self.read_number()

        return self.last_term, self.last_offset, frequency

    def go_to_term(self, term, offset, info_offset):

        """
        Seek past the entry for 'term' having 'offset' to 'info_offset'. This
        permits the scanning for later terms from the specified term.
        """

        self.f.seek(info_offset)
        self.last_term = term
        self.last_offset = offset

class TermIndexWriter(TermWriter):

    "Writing term dictionary index details to files."

    def reset(self):
        TermWriter.reset(self)
        self.last_info_offset = 0

    def write_term(self, term, offset, frequency, info_offset):

        """
        Write the given 'term', its position file 'offset', and its 'frequency'
        to the term dictionary index file, along with the 'info_offset' in the
        term information file.
        """

        TermWriter.write_term(self, term, offset, frequency)

        # Write the information file offset delta.

        self.write_number(info_offset - self.last_info_offset)
        self.last_info_offset = info_offset

class TermIndexReader(TermReader):

    "Reading term dictionary index details from files."

    def reset(self):
        TermReader.reset(self)
        self.last_info_offset = 0

    def read_term(self):

        """
        Read a term, its position file offset, its frequency, and its term
        information file offset from the term dictionary index file.
        """

        term, offset, frequency = TermReader.read_term(self)

        # Read the offset delta.

        self.last_info_offset += self.read_number()

        return term, offset, frequency, self.last_info_offset

class TermDictionaryWriter:

    "Writing term dictionaries."

    def __init__(self, info_writer, index_writer, position_writer, interval):
        self.info_writer = info_writer
        self.index_writer = index_writer
        self.position_writer = position_writer
        self.interval = interval
        self.entry = 0

    def _write_term(self, term, offset, frequency):

        """
        Write the given 'term', its position file 'offset', and its 'frequency'
        to the term information file and optionally to the index, making a
        dictionary entry.
        """

        info_offset = self.info_writer.write_term(term, offset, frequency)

        if self.entry % self.interval == 0:
            self.index_writer.write_term(term, offset, frequency, info_offset)

        self.entry += 1

    def write_term_positions(self, term, doc_positions):

        """
        Write the given 'term' and the 'doc_positions' recording the documents
        and positions at which the term is found.
        """

        offset, frequency = self.position_writer.write_all_positions(doc_positions)
        self._write_term(term, offset, frequency)

    def close(self):
        self.info_writer.close()
        self.index_writer.close()
        self.position_writer.close()

class TermDictionaryReader:

    "Reading term dictionaries."

    def __init__(self, info_reader, index_reader, position_reader):
        self.info_reader = info_reader
        self.index_reader = index_reader
        self.position_reader = position_reader

        self.terms = []
        try:
            while 1:
                self.terms.append(self.index_reader.read_term())
        except EOFError:
            pass

        # Large numbers for ordering purposes.

        self.max_offset = self.terms[-1][1]
        self.max_info_offset = self.terms[-1][2]

    def _find_term(self, term):

        """
        Find the position file offset and frequency of 'term' from the term
        dictionary.
        """

        i = bisect_right(self.terms, (term, self.max_offset, self.max_info_offset)) - 1

        # Get the entry position providing the term or one preceding it.

        if i == -1:
            return None

        found_term, offset, frequency, info_offset = self.terms[i]

        # Where the term is found immediately, return the offset.

        if term == found_term:
            return offset, frequency

        # Otherwise, seek past the index term's entry in the information file
        # and scan for the desired term.

        else:
            self.info_reader.go_to_term(found_term, offset, info_offset)
            try:
                while term > found_term:
                    found_term, offset, frequency = self.info_reader.read_term()
            except EOFError:
                pass

            # If the term is found, return the offset and frequency.

            if term == found_term:
                return offset, frequency
            else:
                return None

    def find_positions(self, term):

        "Return the documents and positions at which the given 'term' is found."

        t = self._find_term(term)
        if t is None:
            return None
        else:
            offset, frequency = t
            return self.position_reader.read_all_positions(offset)

    def get_frequency(self, term):

        "Return the frequency of the given 'term'."

        t = self._find_term(term)
        if t is None:
            return None
        else:
            offset, frequency = t
            return frequency

    def close(self):
        self.info_reader.close()
        self.index_reader.close()
        self.position_reader.close()

# Specific classes for storing document information.

class FieldWriter(FileWriter):

    "Writing field data to files."

    def reset(self):
        self.last_docnum = 0

    def write_fields(self, docnum, fields):

        """
        Write for the given 'docnum', a list of 'fields' (strings representing
        field values). Return the offset at which the fields are stored.
        """

        offset = self.f.tell()

        # Write the document number delta.

        self.write_number(docnum - self.last_docnum)

        # Write the number of fields.

        self.write_number(len(fields))

        # Write the fields themselves.

        for field in fields:
            self.write_string(field, 1) # compress

        self.last_docnum = docnum
        return offset

class FieldReader(FileReader):

    "Reading field data from files."

    def reset(self):
        self.last_docnum = 0

    def read_fields(self):

        """
        Read fields from the file, returning a tuple containing the document
        number and a list of field values.
        """

        # Read the document number.

        self.last_docnum += self.read_number()

        # Read the number of fields.

        nfields = self.read_number()

        # Collect the fields.

        fields = []
        i = 0

        while i < nfields:
            fields.append(self.read_string(1)) # decompress
            i += 1

        return self.last_docnum, fields

    def read_document_fields(self, docnum, offset):

        """
        Read fields for 'docnum' at the given 'offset'. This permits the
        retrieval of details for the specified document, as well as scanning for
        later documents.
        """

        self.f.seek(offset)
        bad_docnum, fields = self.read_fields()
        self.last_docnum = docnum
        return docnum, fields
        
class FieldIndexWriter(FileWriter):

    "Writing field index details to files."

    def reset(self):
        self.last_docnum = 0
        self.last_offset = 0

    def write_document(self, docnum, offset):

        """
        Write for the given 'docnum', the 'offset' at which the fields for the
        document are stored in the fields file.
        """

        # Write the document number and offset deltas.

        self.write_number(docnum - self.last_docnum)
        self.write_number(offset - self.last_offset)

        self.last_docnum = docnum
        self.last_offset = offset

class FieldIndexReader(FileReader):

    "Reading field index details from files."

    def reset(self):
        self.last_docnum = 0
        self.last_offset = 0

    def read_document(self):

        "Read a document number and field file offset."

        # Read the document number delta and offset.

        self.last_docnum += self.read_number()
        self.last_offset += self.read_number()

        return self.last_docnum, self.last_offset

class FieldDictionaryWriter:

    "Writing field dictionary details."

    def __init__(self, field_writer, field_index_writer, interval):
        self.field_writer = field_writer
        self.field_index_writer = field_index_writer
        self.interval = interval
        self.entry = 0

    def write_fields(self, docnum, fields):

        "Write details of the document with the given 'docnum' and 'fields'."

        offset = self.field_writer.write_fields(docnum, fields)

        if self.entry % self.interval == 0:
            self.field_index_writer.write_document(docnum, offset)

        self.entry += 1

    def close(self):
        self.field_writer.close()
        self.field_index_writer.close()

class FieldDictionaryReader:

    "Reading field dictionary details."

    def __init__(self, field_reader, field_index_reader):
        self.field_reader = field_reader
        self.field_index_reader = field_index_reader

        self.docs = []
        try:
            while 1:
                self.docs.append(self.field_index_reader.read_document())
        except EOFError:
            pass

        # Large numbers for ordering purposes.

        self.max_offset = self.docs[-1][1]

    def read_fields(self, docnum):

        "Read the fields of the document with the given 'docnum'."

        i = bisect_right(self.docs, (docnum, self.max_offset)) - 1

        # Get the entry position providing the term or one preceding it.

        if i == -1:
            return None

        found_docnum, offset = self.docs[i]

        # Read from the fields file.

        found_docnum, fields = self.field_reader.read_document_fields(found_docnum, offset)

        # Scan for the document, if necessary.

        try:
            while docnum > found_docnum:
                found_docnum, fields = self.field_reader.read_fields()
        except EOFError:
            pass

        # If the document is found, return the fields.

        if docnum == found_docnum:
            return fields
        else:
            return None

    def close(self):
        self.field_reader.close()
        self.field_index_reader.close()

# High-level classes.

class IndexWriter:

    """
    Building term information and writing it to the term and field dictionaries.
    """

    def __init__(self, dict_writer, field_dict_writer):
        self.dict_writer = dict_writer
        self.field_dict_writer = field_dict_writer
        self.terms = {}
        self.docs = {}

    def add_position(self, term, docnum, position):

        """
        Add a position entry for the given 'term' in the document with the given
        'docnum', indicating the given 'position'.
        """

        if not self.terms.has_key(term):
            doc_positions = self.terms[term] = {}
        else:
            doc_positions = self.terms[term]

        if not doc_positions.has_key(docnum):
            doc = doc_positions[docnum] = []
        else:
            doc = doc_positions[docnum]

        doc.append(position)

    def add_fields(self, docnum, fields):

        "Add for the document with the given 'docnum' a list of 'fields'."

        if not self.docs.has_key(docnum):
            doc_fields = self.docs[docnum] = fields
        else:
            self.docs[docnum] += fields

    def close(self):
        if self.dict_writer is None:
            return

        # Get the terms in order.

        terms = self.terms.items()
        terms.sort()

        for term, doc_positions in terms:
            doc_positions = doc_positions.items()
            doc_positions.sort()
            self.dict_writer.write_term_positions(term, doc_positions)

        self.dict_writer.close()
        self.dict_writer = None

        # Get the documents in order.

        docs = self.docs.items()
        docs.sort()

        for docnum, fields in docs:
            self.field_dict_writer.write_fields(docnum, fields)

        self.field_dict_writer.close()
        self.field_dict_writer = None

class IndexReader:

    "Accessing the term and field dictionaries."

    def __init__(self, dict_reader, field_dict_reader):
        self.dict_reader = dict_reader
        self.field_dict_reader = field_dict_reader

    def find_positions(self, term):
        return self.dict_reader.find_positions(term)

    def get_frequency(self, term):
        return self.dict_reader.get_frequency(term)

    def get_fields(self, docnum):
        return self.field_dict_reader.read_fields(docnum)

    def close(self):
        self.dict_reader.close()
        self.field_dict_reader.close()

class Index:

    "An inverted index solution encapsulating the various components."

    def __init__(self, pathname):
        self.pathname = pathname
        self.reader = None
        self.writer = None

    def get_writer(self, interval=INTERVAL):

        "Return a writer, optionally using the given indexing 'interval'."

        if not exists(self.pathname):
            mkdir(self.pathname)

        tdf = open(join(self.pathname, "terms"), "wb")
        info_writer = TermWriter(tdf)

        tdif = open(join(self.pathname, "index"), "wb")
        index_writer = TermIndexWriter(tdif)

        tpf = open(join(self.pathname, "positions"), "wb")
        positions_writer = PositionWriter(tpf)

        dict_writer = TermDictionaryWriter(info_writer, index_writer, positions_writer, interval)

        ff = open(join(self.pathname, "fields"), "wb")
        field_writer = FieldWriter(ff)

        fif = open(join(self.pathname, "fields_index"), "wb")
        field_index_writer = FieldIndexWriter(fif)

        field_dict_writer = FieldDictionaryWriter(field_writer, field_index_writer, interval)

        self.writer = IndexWriter(dict_writer, field_dict_writer)
        return self.writer

    def get_reader(self):

        "Return a reader for the index."

        if not exists(self.pathname):
            raise OSError, "Index path %r does not exist." % self.pathname

        tdf = open(join(self.pathname, "terms"), "rb")
        info_reader = TermReader(tdf)

        tdif = open(join(self.pathname, "index"), "rb")
        index_reader = TermIndexReader(tdif)

        tpf = open(join(self.pathname, "positions"), "rb")
        positions_reader = PositionReader(tpf)

        dict_reader = TermDictionaryReader(info_reader, index_reader, positions_reader)

        ff = open(join(self.pathname, "fields"), "rb")
        field_reader = FieldReader(ff)

        fif = open(join(self.pathname, "fields_index"), "rb")
        field_index_reader = FieldIndexReader(fif)

        field_dict_reader = FieldDictionaryReader(field_reader, field_index_reader)

        self.reader = IndexReader(dict_reader, field_dict_reader)
        return self.reader

    def close(self):
        if self.reader is not None:
            self.reader.close()
            self.reader = None
        if self.writer is not None:
            self.writer.close()
            self.writer = None

# vim: tabstop=4 expandtab shiftwidth=4
